using System;
using System.Collections.Generic;

namespace Syntriax.Engine.Core;

public class ActiveBehaviourCollector<T> : IBehaviourCollector<T> where T : class, IBehaviour
{
    public Event<IBehaviourCollector<T>, IBehaviourCollector<T>.BehaviourCollectedArguments> OnCollected { get; } = new();
    public Event<IBehaviourCollector<T>, IBehaviourCollector<T>.BehaviourRemovedArguments> OnRemoved { get; } = new();
    public Event<IHasUniverse> OnUniverseAssigned { get; } = new();
    public Event<IAssignable>? OnUnassigned { get; } = new();

    private readonly Event<IBehaviourController, IBehaviourController.BehaviourAddedArguments>.EventHandler cachedOnBehaviourAdded = null!;
    private readonly Event<IBehaviourController, IBehaviourController.BehaviourRemovedArguments>.EventHandler cachedOnBehaviourRemoved = null!;
    private readonly Event<IActive, IActive.ActiveChangedArguments>.EventHandler cachedOnBehaviourStateChanged = null!;
    private readonly Event<IUniverse, IUniverse.UniverseObjectRegisteredArguments>.EventHandler cachedOnUniverseObjectRegistered = null!;
    private readonly Event<IUniverse, IUniverse.UniverseObjectUnRegisteredArguments>.EventHandler cachedOnUniverseObjectUnregistered = null!;

    private readonly List<T> monitoringBehaviours = new(32);
    protected readonly List<T> activeBehaviours = new(32);
    protected readonly Dictionary<IActive, T> monitoringActiveToBehaviour = new(32);

    public IUniverse Universe { get; private set; } = null!;

    private void OnUniverseObjectRegistered(IUniverse manager, IUniverse.UniverseObjectRegisteredArguments arguments)
    {
        IUniverseObject universeObject = arguments.UniverseObjectRegistered;

        universeObject.BehaviourController.OnBehaviourAdded.AddListener(cachedOnBehaviourAdded);
        universeObject.BehaviourController.OnBehaviourRemoved.AddListener(cachedOnBehaviourRemoved);

        for (int i = 0; i < universeObject.BehaviourController.Count; i++)
            OnBehaviourAdded(universeObject.BehaviourController, new(universeObject.BehaviourController[i]));
    }

    private void OnUniverseObjectUnregistered(IUniverse manager, IUniverse.UniverseObjectUnRegisteredArguments arguments)
    {
        IUniverseObject universeObject = arguments.UniverseObjectUnregistered;

        universeObject.BehaviourController.OnBehaviourAdded.RemoveListener(cachedOnBehaviourAdded);
        universeObject.BehaviourController.OnBehaviourRemoved.RemoveListener(cachedOnBehaviourRemoved);

        for (int i = 0; i < universeObject.BehaviourController.Count; i++)
            OnBehaviourRemoved(universeObject.BehaviourController, new(universeObject.BehaviourController[i]));
    }

    protected virtual void OnBehaviourAdd(IBehaviour behaviour) { }
    private void OnBehaviourAdded(IBehaviourController controller, IBehaviourController.BehaviourAddedArguments arguments)
    {
        if (arguments.BehaviourAdded is not T tBehaviour)
            return;

        monitoringBehaviours.Add(tBehaviour);
        monitoringActiveToBehaviour.Add(tBehaviour, tBehaviour);
        tBehaviour.OnActiveChanged.AddListener(cachedOnBehaviourStateChanged);
        OnBehaviourStateChanged(tBehaviour, new(!tBehaviour.IsActive));
    }

    private void OnBehaviourStateChanged(IActive sender, IActive.ActiveChangedArguments arguments)
    {
        T behaviour = monitoringActiveToBehaviour[sender];
        if (sender.IsActive)
        {
            activeBehaviours.Add(behaviour);
            OnBehaviourAdd(behaviour);
            OnCollected?.Invoke(this, new(behaviour));
        }
        else if (activeBehaviours.Remove(behaviour))
        {
            OnBehaviourRemove(behaviour);
            OnRemoved?.Invoke(this, new(behaviour));
        }
    }

    protected virtual void OnBehaviourRemove(IBehaviour behaviour) { }
    private void OnBehaviourRemoved(IBehaviourController controller, IBehaviourController.BehaviourRemovedArguments arguments)
    {
        if (arguments.BehaviourRemoved is not T tBehaviour)
            return;

        if (!monitoringBehaviours.Remove(tBehaviour) || !monitoringActiveToBehaviour.Remove(tBehaviour))
            return;

        tBehaviour.OnActiveChanged.RemoveListener(cachedOnBehaviourStateChanged);
        if (activeBehaviours.Remove(tBehaviour))
        {
            OnBehaviourRemove(tBehaviour);
            OnRemoved?.Invoke(this, new(tBehaviour));
        }
    }

    public bool Assign(IUniverse universe)
    {
        if (Universe is not null)
            return false;

        foreach (IUniverseObject universeObject in universe.UniverseObjects)
            OnUniverseObjectRegistered(universe, new(universeObject));

        universe.OnUniverseObjectRegistered.AddListener(cachedOnUniverseObjectRegistered);
        universe.OnUniverseObjectUnRegistered.AddListener(cachedOnUniverseObjectUnregistered);

        Universe = universe;
        OnUniverseAssigned?.Invoke(this);

        return true;
    }

    public bool Unassign()
    {
        if (Universe is null)
            return false;

        foreach (IUniverseObject universeObject in Universe.UniverseObjects)
            OnUniverseObjectUnregistered(Universe, new(universeObject));

        Universe.OnUniverseObjectRegistered.RemoveListener(cachedOnUniverseObjectRegistered);
        Universe.OnUniverseObjectUnRegistered.RemoveListener(cachedOnUniverseObjectUnregistered);

        Universe = null!;
        OnUnassigned?.Invoke(this);
        return true;
    }

    public int Count => activeBehaviours.Count;
    public T this[Index index] => activeBehaviours[index];

    public ActiveBehaviourCollector()
    {
        cachedOnBehaviourAdded = OnBehaviourAdded;
        cachedOnBehaviourRemoved = OnBehaviourRemoved;
        cachedOnBehaviourStateChanged = OnBehaviourStateChanged;
        cachedOnUniverseObjectRegistered = OnUniverseObjectRegistered;
        cachedOnUniverseObjectUnregistered = OnUniverseObjectUnregistered;
    }
}
